// "newactual.C"
// Actual objects to be instantiated.

/*    Copyright (C) 1994,1995,1996  Steve Hardt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Steve Hardt 
    satan@xevil.com
    http://www.xevil.com
*/

#ifndef NO_PRAGMAS
#pragma implementation "newactual.h"
#endif

#include <strstream.h>
#include "utils.h"
#include "coord.h"
#include "world.h"
#include "physical.h"
#include "actual.h"
#include "newactual.h"

// Bitmap files.  Also contain constants and some simple methods.
// Most importantly, they contain the "contexts".
#include "bitmaps/bazooka/bitmaps"
#include "bitmaps/rocket/bitmaps"
#include "bitmaps/blood/bitmaps"
#include "bitmaps/oilblood/bitmaps"
#include "bitmaps/alienblood/bitmaps"
#include "bitmaps/doppel/bitmaps"
#include "bitmaps/transdoppel/bitmaps"
#include "bitmaps/vendor/bitmaps"
#include "bitmaps/smoke/bitmaps"
#include "bitmaps/scud/bitmaps"
#include "bitmaps/humantorch/bitmaps"
#include "bitmaps/flameon/bitmaps"
#include "bitmaps/fire/bitmaps"
#include "bitmaps/firebomb/bitmaps"
#include "bitmaps/firemaker/bitmaps"
#include "bitmaps/firebombs/bitmaps"
#include "bitmaps/rockman/bitmaps"
#include "bitmaps/rockman/buster/bitmaps"
#include "bitmaps/rockman/death/bitmaps"
#include "bitmaps/rockman/enter/bitmaps"
#include "bitmaps/k_shield/bitmaps"
#include "bitmaps/r_shield/bitmaps"
#include "bitmaps/k_protection/bitmaps"
#include "bitmaps/r_protection/bitmaps"
#include "bitmaps/jetpack/bitmaps"
#include "bitmaps/kenny/bitmaps"
#include "bitmaps/swastikarang/bitmaps"
#include "bitmaps/swastikarangs/bitmaps"

/*************Bazooka*************/


Bazooka::Bazooka(WorldP w,LocatorP l,const Pos &p) 
: Gun(context,xdata,w,l,p) 
{}



Size Bazooka::get_shot_size(Dir dir)
{
  return Rocket::get_size(dir);
}



PhysicalP Bazooka::create_shot(PhysicalP shooter,WorldP world,
				LocatorP locator,const Pos &pos,Dir dir)
{
  PhysicalP shot = new Rocket(world,locator,pos,shooter->get_id(),dir);
  assert (shot);
  return shot;
}


//*********Bazooka end***********/


/*************Rocket*******************/

Rocket::Rocket(WorldP w,LocatorP l,const Pos &p,const Id &sh,Dir d)
: Shot(context,xdata,w,l,p,sh,d,d) 
{
  stats.add_creation();
  armingTimer.set(ROCKET_ARMING_TIME);
}


int Rocket::get_anim_time()
{
   return 1;
}


void Rocket::die()
{
   WorldP world = get_world();
   LocatorP locator = get_locator();
   const Area area = get_area();
  
   if (armingTimer.ready()) {
      PhysicalP explosion = 
         new Explosion(world,locator,area.get_middle(),shooter,
         ROCKET_EXPLOSION_RADIUS,ROCKET_EXPLOSION_DAMAGE_MAX);
      assert (explosion);
      locator->add(explosion);
   }

   Shot::die();
}

void Rocket::act()
{
  LocatorP locator = get_locator();
  WorldP world = get_world();
  const Area area = get_area();
      
  PhysicalP smoke = 
     new Smoke(world,locator,area.get_middle());
  assert (smoke);
  locator->add(smoke);

  armingTimer.clock();

  Shot::act();
}


Stats Rocket::stats;

/*************Rocket end*********/


/*************Begin Blood stuff****/

BloodSplat::BloodSplat(const FallingContext &fc, FallingXdata &fx, WorldP w,LocatorP l,const Pos &pos, Dir dir,Speed speed)
  :Falling(fc, fx, w, l, pos, dir)
{
  assert(dir != CO_air);
  const Vel *unitVels = get_unit_vels();
  set_vel_next(speed * unitVels[dir]);

  splat=False;

  timer.set(BLOOD_TIME);
}

BloodSplat::BloodSplat(const FallingContext &fc, FallingXdata &fx, WorldP w, LocatorP l,const Pos &pos, const Vel &vel)
  :Falling(fc, fx, w, l, pos, vel.get_dir())
{
  splat=False;

  set_vel_next(vel);
  timer.set(BLOOD_TIME);
}

Boolean BloodSplat::collidable()
{
  return False;
}


void BloodSplat::act()
{
  WorldP world = get_world();
  const Area &area = get_area();
  Touching touching;
  Hanging hanging;

  world->touchingHanging(touching, hanging, area);

  if (hit_wall() && !splat) {
     splat=True;
     set_vel_next(0);
     if (touching == CO_dn) {
        set_dir_next(CO_center);
     } else if (touching == CO_l) {
        set_dir_next(CO_l);
     } else if (touching == CO_r) {
        set_dir_next(CO_r);
     }
     set_falling(False);
  }

  if (timer.ready())
    kill_self();

  timer.clock();

  Falling::act();
}


/*******Normal bio-blood********/


Blood::Blood(WorldP w,LocatorP l,const Pos &pos,
		 Dir dir,Speed speed)
: BloodSplat(context, xdata, w, l, pos, dir, speed)
{
}



Blood::Blood(WorldP w,LocatorP l,const Pos &pos,
		 const Vel &vel)
: BloodSplat(context, xdata, w, l, pos, vel)
{
}


Stats Blood::stats;


/**********Machines' oil 'blood'********/


OilBlood::OilBlood(WorldP w,LocatorP l,const Pos &pos,
		 Dir dir,Speed speed)
: BloodSplat(context, xdata, w, l, pos, dir, speed)
{
}



OilBlood::OilBlood(WorldP w,LocatorP l,const Pos &pos,
		 const Vel &vel)
: BloodSplat(context, xdata, w, l, pos, vel)
{
}

Stats OilBlood::stats;


/**********Aliens' green blood.  Optionally this could be modified to implement acid-burn.**********/


AlienBlood::AlienBlood(WorldP w,LocatorP l,const Pos &pos,
		 Dir dir,Speed speed)
: BloodSplat(context, xdata, w, l, pos, dir, speed)
{
}



AlienBlood::AlienBlood(WorldP w,LocatorP l,const Pos &pos,
		 const Vel &vel)
: BloodSplat(context, xdata, w, l, pos, vel)
{
}

Stats AlienBlood::stats;


/*************End blood stuff*****************/


Doppeloid::Doppeloid(const AutoUseContext &ac,
	AutoUseXdata &aux, WorldP w,LocatorP l,const Pos &pos)
: AutoUse(ac, aux, w, l, pos)
{}



void Doppeloid::use(PhysicalP p)
{
  assert(p && p->get_intel());

  IntelP userintel = p->get_intel();
  ostrstream bitchname;
  bitchname << userintel->get_name() << "'s bitch" << ends;

  // If p is not allowed to use doppelgangers, just return.
  const PhysicalContext *usersPC = p->get_context();
  if (!usersPC->doppelUser)
    return;

  WorldP world = get_world();
  LocatorP locator = get_locator();
  
  PhysicalP obj = create_physical(p->get_area(),p);

  IntelOptions ops;
  ops.psychotic = True;
  ops.classFriends = False;
  ops.ignoreItems = True;
  ops.ignoreLemmings = True;
  NeutralP neutral = 
//    new Neutral(world,locator,"Slave",&ops,
    new Neutral(world,locator,bitchname.str(),&ops,
		ITpsychotic|ITclassFriends|ITignoreItems|ITignoreLemmings,
		p->get_intel());
  assert(neutral);
  locator->register_neutral(neutral);

  locator->add(obj);
  obj->set_intel(neutral);
  obj->set_dont_collide(p);
  
  kill_self();
//  Item::use(p);
  AutoUse::use(p);
}


PhysicalP Doppeloid::create_physical(const Area &area,PhysicalP original)
{
  WorldP world = get_world();
  LocatorP locator = get_locator();
  Pos middle = area.get_middle();
  PhysicalP obj;
  
  // Get the PhysicalContext(description of its class) of the original
  const PhysicalContext *originalContext = original->get_context();

  // Use the original's PhysicalContext to create a new object.
  obj = originalContext->create(world,locator,
				middle - 0.5 * originalContext->get_size_max());

  assert(obj);
  return obj;
}


Doppel::Doppel(WorldP w,LocatorP l,const Pos &pos)
: Doppeloid(context, xdata, w, l, pos)
{}

void Doppel::use(PhysicalP p)
{
   stats.add_use();
   Doppeloid::use(p);
}

Stats Doppel::stats;


/**Transmorphic Doppel*/


TransDoppel::TransDoppel(WorldP w,LocatorP l,const Pos &pos)
: Doppeloid(context, xdata, w, l, pos)
{}


void TransDoppel::use(PhysicalP p)
{
   stats.add_use();
   Doppeloid::use(p);
}

// New version which uses Locator::filter_contexts.
// More general and easy to extend.
PhysicalP TransDoppel::create_physical(const Area &area,PhysicalP original)
{
  Pos middle = area.get_middle();

  const PhysicalContext *list[A_CLASSES_NUM];
  int size;
  LocatorP l = get_locator();
  WorldP w = get_world();

  // Get list of all classes that are potential transmogify targets.
  size = l->filter_contexts(list,NULL,transmogify_target_filter);

  // Something is wrong if size == 0, because there are no classes to 
  // transmogify into.
  assert(size);
  
  // Randomly choose.
  int which = Utils::choose(size);

  // Create object centered on middle.
  Pos pos = middle - 0.5 * list[which]->get_size_max();
  PhysicalP obj = list[which]->create(w,l,pos);

  return obj;
}


Boolean TransDoppel::transmogify_target_filter(const PhysicalContext *pc)
{
  return pc->transmogifyTarget;
}


Stats TransDoppel::stats;


/**************Vendor************/

ScudCounter::ScudCounter()
{
//   counter = Utils::choose(1400);
   init();
}

void ScudCounter::init()
{
   Utils::randomList(numberpool, numbermax);
   numberused = numbermax-1;
//It -is- possible to get Scud #0.
}

int ScudCounter::assign_number()
{
   int temp;

   temp = peek_number();
   numberused--;

   if (numberused < 0)
      init();

   return temp;
}


int ScudCounter::peek_number()
{
   return numberpool[numberused];
}


Vendor::Vendor(WorldP w,LocatorP l,const Pos &pos)
: AutoUse(context,xdata,w,l,pos)
{}


void Vendor::use(PhysicalP p)
{
  assert(p && p->get_intel());

  ostrstream scudName;

  scudName << "Heart Breaker Series " << scudCounter.peek_number() << ends;

  // If p is not allowed to use doppelgangers, just return.
  const PhysicalContext *usersPC = p->get_context();
  if (!usersPC->doppelUser)
    return;

  stats.add_use();
  WorldP world = get_world();
  LocatorP locator = get_locator();
  
  UserP obj = create_scud(p->get_area(),p);

  IntelOptions ops;
  ops.psychotic = True;
  ops.classFriends = False;
  ops.ignoreItems = True;
  ops.ignoreLemmings = True;
  NeutralP neutral = 
    new Neutral(world,locator,scudName.str(),&ops,
		ITpsychotic|ITclassFriends|ITignoreItems|ITignoreLemmings,
		p->get_intel());
  assert(neutral);
  locator->register_neutral(neutral);

  locator->add(obj);
  obj->set_intel(neutral);
  obj->set_dont_collide(p);

  //Try to give newborn scud a gun.

  Pistol *scudpistol = new Pistol(world, locator, p->get_area().get_middle());
  locator->add(scudpistol);
  obj->give_weapon(scudpistol);

  kill_self();
  AutoUse::use(p);
  set_quiet_death();

  ostrstream msg;
  msg << "ATTENTION: Scud disposable robot assassain, heartbreaker series "
	<< scudCounter.assign_number() << endl
	<< "This unit will self-destruct upon termination of target." << ends;
  locator->message_enq(msg.str());

}


UserP Vendor::create_scud(const Area &area,PhysicalP original)
{
  WorldP world = get_world();
  LocatorP locator = get_locator();
  Pos middle = area.get_middle();
  Scud *obj;
  
  // Use the original's PhysicalContext to create a new object.
  obj = new Scud(world,locator,
			middle - 0.5 * Scud::get_size_max());

  assert(obj);
  return obj;
}


Stats Vendor::stats;

ScudCounter Vendor::scudCounter;

/************SCUD*************/

Scud::Scud(WorldP w,LocatorP l,const Pos &rawPos) 
     : Creature(creatureContext,creatureXdata,w,l,rawPos),
       Walking(walkingContext,walkingXdata),
       Fighter(fighterContext,fighterXdata),
       User(userContext,userXdata)
{}

Boolean Scud::is_fighter() 
{
  return True;
}


Boolean Scud::is_walking() 
{
  return True;
}



Boolean Scud::is_user() 
{
  return True;
}



void Scud::get_pixmap_mask(int dpyNum,Pixmap &pixmap,Pixmap &mask,Dir dir,int animNum)
{
  Fighter::get_pixmap_mask(dpyNum,pixmap,mask,dir,animNum);
}



void Scud::get_size_offset_next(Size &size,Size &offset,Dir dirNext)
{
  Fighter::get_size_offset_next(size,offset,dirNext);
}



int Scud::get_weapons_num()
{
  return User::get_weapons_num();
}



int Scud::get_items_num()
{
  return User::get_items_num();
}



PhysicalP Scud::get_weapon(int n)
{
  return User::get_weapon(n);
}



PhysicalP Scud::get_item(int n)
{
  return User::get_item(n);
}




PhysicalP Scud::get_weapon_current()
{
  return User::get_weapon_current();
}



PhysicalP Scud::get_item_current()
{
  return User::get_item_current();
}



int Scud::get_item_count()
{
  return User::get_item_count();
}



void Scud::set_mapped_next(Boolean val)
{
  User::set_mapped_next(val);
}



ClassId Scud::get_weapon_string(const char *&str)
{
  return User::get_weapon_string(str);
}



void Scud::act()
{
  Walking::_act();
  Fighter::_act();
  Creature::act();
  User::_act();
  // NOTE: Creature::act() before User::act();
}



void Scud::update()
{
  // Sticky::_update();
  Fighter::_update();
  User::_update();
  Creature::update();
}



void Scud::die()
{
  if (!get_quiet_death())
    stats.add_death(get_birth_time());

  User::_die();
  Creature::die();


  WorldP world = get_world();
  LocatorP locator = get_locator();
  const Area area = get_area();
  
  if (!get_quiet_death()) {
     PhysicalP explosion = 
     new Explosion(world,locator,area.get_middle(), get_id(),
	      SCUD_EXPLOSION_RADIUS,SCUD_EXPLOSION_DAMAGE_MAX);
     assert (explosion);
     locator->add(explosion);
  }
}



void Scud::collide(PhysicalP other) 
{
  if (!Fighter::_collide(other) && !User::_collide(other))
    Creature::collide(other);
}



void Scud::init_x(Xvars &xvars)
{
  Fighter::init_x(xvars);
}



Stats Scud::stats;



/**********END Scud***********/


/***********Smoke********/
Smoke::Smoke(WorldP w,LocatorP l,const Pos &mid) : 
  Moving(context,xdata,w,l,uncenter(mid))
{
  timer.set(3);
}


Pos Smoke::uncenter(Pos mid)
{
  Size maxsize = get_size();

  Pos pos;
  pos.x = mid.x - maxsize.width/2;
  pos.y = mid.y - maxsize.height/2;

  return pos;
}


int Smoke::get_anim_time() {
   return 0;
}


Boolean Smoke::ignore_walls()
{
   return True;
}


Boolean Smoke::collidable()
{
  return False;
}


int Smoke::get_drawing_level()
{
  return 2;
}


void Smoke::act() 
{
  timer.clock();
  if (timer.ready()) 
     kill_self();
  Moving::act();
}

/*
void Smoke::update()
{
  area = areaNext;
  Physical::update();
}
*/


/**********End Smoke***********/


/**********Human Torch********/

HumanTorch::HumanTorch(WorldP w, LocatorP l, const Pos &pos)
   : Shield(context, xdata, w, l, pos)
{}

void HumanTorch::use(PhysicalP user) {
   LocatorP locator = get_locator();
   ostrstream msg;
   
   msg << user->get_intel()->get_name() << ": FLAME ON!" << ends;
   locator->message_enq(msg.str());

   stats.add_use();

   Shield::use(user);
   set_quiet_death();
}


ProtectionP HumanTorch::create_protection(const Area &area)
{
   ProtectionP pr = new FlameOn(get_world(),get_locator(),area);
   assert(pr);
   return pr;
}

Stats HumanTorch::stats;


FlameOn::FlameOn(WorldP w, LocatorP l, const Area &area)
   :Protection(context, xdata, w, l, area)
{
//   flamearea = area;
}

int FlameOn::corporeal_defend(int damage)
{
   return damage/2;
}

int FlameOn::heat_defend(int heat, Boolean secondary)
{
   return 0;
}

void FlameOn::draw(Drawable, Xvars &, int, const Area &)
{}

void FlameOn::act()
{
   Pos pos;
   Size size;
   const Area &area = get_area();

   area.get_rect(pos,size);
   pos.x += Utils::choose(size.width);
   // Start fire in upper-half.
   if (size.height > 1)
     pos.y += Utils::choose((int)(size.height * .5)); 

   LocatorP locator = get_locator();
   PhysicalP fire = new Fire(get_world(),locator,pos);
   fire->set_dont_collide(this);
   locator->add(fire);
}






Fire::Fire(WorldP w,LocatorP l,const Pos &mid,Boolean coll)
 :  NewAnimated(context, xdata, w, l, uncenter(mid))
{
   Vel vel(0, FIRE_VEL_Y);

   frame = 0;

   isCollidable = coll;

   set_vel(vel);
   set_vel_next(vel);

   smokeTimer.set(3);

   makeSmoke();

}


int Fire::get_drawing_level()
{
   return 2;
}


int Fire::get_damage() 
{
  return FIRE_DAMAGE;
}




Boolean Fire::collidable()
{
  return isCollidable;
}



Boolean Fire::ignore_walls()
{
  return True;
}


void Fire::avoid(PhysicalP)
{}



void Fire::collide(PhysicalP other)
{
  if (other->is_shot())
    return;
  other->heat_attack(NULL,FIRE_HEAT,True);
}



void Fire::corporeal_attack(PhysicalP, int)
{}



void Fire::heat_attack(PhysicalP, int, Boolean)
{}


void Fire::act()
{
  if (smokeTimer.ready()) {
     makeSmoke();
     smokeTimer.set(3);
  }

  if (frame >= FIRE_FRAME_LAST) {
     kill_self();
  } else {
     set_frame_next(++frame);
  }

  if (Utils::choose(8)>6) {
     makeSmoke();
  }

  NewAnimated::act();
}


void Fire::makeSmoke()
{
   /**"Where there's fire there's smoke", or something.  */
   LocatorP locator = get_locator();
   WorldP world = get_world();
   const Area smokearea = get_area();
   Pos pos;
   Size size;
   Vel smokevel;

   smokevel.dy=FIRE_VEL_Y*2;
   smokevel.dx=0;

   smokearea.get_rect(pos, size);
   pos.x += Utils::choose(size.width);
   pos.y += Utils::choose(size.height / 2);

   Smoke *smokepuff = new Smoke(world, locator, pos);
   smokepuff->set_vel_next(smokevel);

   assert (smokepuff);
   locator->add(smokepuff);

   /***********************************/

}



Pos Fire::uncenter(Pos mid)
{
  Size maxsize = get_size_max();

  Pos pos;
  pos.x = mid.x - maxsize.width/2;
  pos.y = mid.y - maxsize.height/2;

  return pos;
}



FireBomb::FireBomb(WorldP w,LocatorP l,const Pos &pos,const Id &sh,
		 Dir dir,Speed speed)
: Falling(context,xdata,w,l,pos,dir)
{
  assert(dir != CO_air);
  const Vel *unitVels = get_unit_vels();
  set_vel_next(speed * unitVels[dir]);
  
  timer.set(FIREBOMB_TIME);
  shooter = sh;
  defused = False;
}



FireBomb::FireBomb(WorldP w,LocatorP l,const Pos &pos,const Id &sh,
		 const Vel &vel)
: Falling(context,xdata,w,l,pos,vel.get_dir())
{
  stats.add_creation();

  set_vel_next(vel);
  timer.set(FIREBOMB_TIME);
  shooter = sh;
  defused = False;
}



void FireBomb::set_quiet_death()
{
  defused = True;
  Falling::set_quiet_death();
}



void FireBomb::act()
{
  if (timer.ready())
    kill_self();
  
  timer.clock();
  Falling::act();
}



void FireBomb::die()
{
  if (!defused)
    {
      WorldP world = get_world();
      LocatorP locator = get_locator();
      const Area area = get_area();
      Pos p = area.get_middle();
      p.y += get_size_max().height/2;
      p.y -= Fireball::get_size_max().height;

      PhysicalP flamer = new FireMaker (shooter, world, locator, p);

      assert (flamer);
      locator->add(flamer);
    }
  Falling::die();
}



Stats FireBomb::stats;



FireMaker::FireMaker(Id sh, WorldP w, LocatorP l, const Pos &p)
: Physical (context, w, l)
{
   shooterId = sh;
   pos = p;
   Size size;
   size.width = size.height = 0;

   Area narea(AR_RECT, pos, size);
   area = narea;
   timer.set(num_to_make);
}


void FireMaker::draw(Drawable, Xvars &, int, const Area &)
{}


Boolean FireMaker::collidable()
{
   return False;
}


const Area &FireMaker::get_area()
{
   return area;
}


const Area &FireMaker::get_area_next()
{
   return area;
}


void FireMaker::act()
{
   WorldP w = get_world();
   LocatorP l = get_locator();

   timer.clock();
   if (timer.ready()) {
      kill_self();
   }

   PhysicalP p1, p2;

   p1 = new Fireball(w, l, pos, shooterId, CO_R);
   p2 = new Fireball(w, l, pos, shooterId, CO_L);

   assert(p1);
   assert(p2);

   l->add(p1);
   l->add(p2);
}



FireBombs::FireBombs(WorldP w,LocatorP l,const Pos &p) 
: Gun(context,xdata,w,l,p) 
{}



Size FireBombs::get_shot_size(Dir dir)
{
  return FireBomb::get_size(dir);
}



PhysicalP FireBombs::create_shot(PhysicalP shooter,WorldP world,
				LocatorP locator,const Pos &pos,Dir dir)
{
  Speed speed;
  if (dir == CO_UP_L || dir == CO_UP || dir == CO_UP_R)
    speed = FIREBOMBS_TOSS_SPEED;
  else if (dir == CO_R || dir == CO_DN_R || dir == CO_DN_L || dir == CO_L)
    speed = FIREBOMBS_ROLL_SPEED;
  else if (dir == CO_DN)
    speed = 0;
  else
    assert(0);

  assert(shooter->is_moving());
  const Vel *unitVels = get_unit_vels();
  Vel vel = ((MovingP)shooter)->get_vel() + speed * unitVels[dir];

  PhysicalP shot = new FireBomb(world,locator,pos,shooter->get_id(),vel);
  assert (shot);
  return shot;
}


RockMan::RockMan(WorldP w,LocatorP l,const Pos &rawPos) 
: Creature(creatureContext,creatureXdata,w,l,rawPos),
  Walking(walkingContext,walkingXdata),
  User(userContext,userXdata),
  BuiltIn(builtInContext,builtInXdata)
{}



Boolean RockMan::is_walking()
{
  return True;
}



Boolean RockMan::is_user()
{
  return True;
}



Boolean RockMan::is_built_in()
{
  return True;
}



int RockMan::get_weapons_num()
{
  return User::get_weapons_num();
}



int RockMan::get_items_num()
{
  return User::get_items_num();
}



PhysicalP RockMan::get_weapon(int n)
{
  return User::get_weapon(n);
}



PhysicalP RockMan::get_item(int n)
{
  return User::get_item(n);
}



PhysicalP RockMan::get_weapon_current()
{
  return User::get_weapon_current();
}



PhysicalP RockMan::get_item_current()
{
  return User::get_item_current();
}



int RockMan::get_item_count()
{
  return User::get_item_count();
}



void RockMan::set_mapped_next(Boolean val)
{
  User::set_mapped_next(val);
}



ClassId RockMan::get_weapon_string(const char *&str)
{
  // I'd prefer to have this functionality higher in the class hierarchy.
  ClassId id = User::get_weapon_string(str);
  
  if (id == A_None)
    return BuiltIn::get_weapon_string(str);
  else
    return id;
}



void RockMan::act()
{
  Walking::_act();
  BuiltIn::_act();
  Creature::act();
  User::_act();
  // NOTE: Creature::act() before User::act();
}



void RockMan::update()
{
  // Walking::_update();
  User::_update();
  // BuiltIn::_update();
  Creature::update();
}



void RockMan::die()
{
  LocatorP locator = get_locator();
  const int spd=5, diag=7;
  int deathidx[12][2] = {{spd, 0}, {0, spd}, {-spd, 0}, {0, -spd},
		{diag, diag}, {diag, -diag}, {-diag, diag}, {-diag, -diag},
		{spd*2, 0}, {0, spd*2}, {-spd*2, 0}, {0,-spd*2}};
  int timeidx[12] = {40, 40, 40, 40, 20, 20, 20, 20, 20, 20, 20, 20};
  MovingP Deathtemp;
  Vel tempVel;
  int cc;
  if (!get_quiet_death()) {
    stats.add_death(get_birth_time());
    for (cc=0; cc<12; cc++) {
       tempVel.dx = deathidx[cc][0];
       tempVel.dy = deathidx[cc][1];
       Deathtemp = new RockDeath(get_world(), get_locator(), 
				get_area().get_middle(), timeidx[cc]);
       assert(Deathtemp);
       Deathtemp->set_vel_next(tempVel);
       locator->add(Deathtemp);
    }
  }

  // Walking::_die();
  User::_die();
  // BuiltIn::_die();
  Creature::die();
}



void RockMan::collide(PhysicalP other)
{
  User::collide(other);
}



int RockMan::get_anim_time()
{
  return ROCKMAN_ANIM_TIME;
}



Dir RockMan::compute_weapon_dir(ITcommand command)
{
  return Buster::compute_weapon_dir(command);
}



Size RockMan::get_shot_size(Dir dir)
{
  return Laser::get_size(dir);
}



PhysicalP RockMan::create_shot(const Pos &pos,Dir dir)
{
  PhysicalP shot =
    new Buster(get_world(), get_locator(), pos, get_id(), dir);
  assert(shot);
  return shot;
}



Stats RockMan::stats;



Buster::Buster(WorldP w,LocatorP l,const Pos &p,const Id &shooter,Dir d)
       : Shot(context,xdata,w,l,p,shooter,d,d) 
{
  stats.add_creation();
}



Dir Buster::compute_weapon_dir(ITcommand command)
{
  // Kind of a hack.
  switch (command) {
  case IT_WEAPON_R:
    return CO_R;
  case IT_WEAPON_DN_R:
    return CO_R;
  case IT_WEAPON_DN:
    return CO_R;
  case IT_WEAPON_DN_L:
    return CO_L;
  case IT_WEAPON_L:
    return CO_L;
  case IT_WEAPON_UP_L:
    return CO_L;
  case IT_WEAPON_UP:
    return CO_L;
  case IT_WEAPON_UP_R:
    return CO_R;
  };
  return CO_air;
}


void Buster::heat_attack(PhysicalP,int,Boolean)
{}


Stats Buster::stats;


RockDeath::RockDeath(WorldP w,LocatorP l,const Pos &mid, int time) : 
  Moving(context,xdata,w,l,uncenter(mid))
{
  timer.set(time);
}


Pos RockDeath::uncenter(Pos mid)
{
  Size maxsize = get_size();

  Pos pos;
  pos.x = mid.x - maxsize.width/2;
  pos.y = mid.y - maxsize.height/2;

  return pos;
}


int RockDeath::get_anim_time() {
   return 0;
}


Boolean RockDeath::ignore_walls()
{
   return True;
}


Boolean RockDeath::collidable()
{
  return False;
}


int RockDeath::get_drawing_level()
{
  return 2;
}


void RockDeath::act() 
{
  timer.clock();
  if (timer.ready()) 
     kill_self();
  Moving::act();
}







RockEnter::RockEnter(WorldP w,LocatorP l,const Pos &topleft) : 
  NewAnimated(context,xdata,w,l,relocate(topleft))
{
   target = topleft;
   Size productSize = RockMan::get_size_max();
   target.x += productSize.width/2;
   target.y += productSize.height/2;
   Vel v(0, 24);
   set_vel_next(v);
   onTarget = False;

  ostrstream msg;
  msg << "                            READY" << ends;
  l->message_enq(msg.str());

}



Pos RockEnter::relocate(Pos topleft)
{
  Size maxsize = get_size();

  Pos pos;
  pos = topleft;
//  Size targetSize = RockMan::get_size_max();
//  pos.x += targetSize.width/2;
  pos.y = 0;

  return pos;
}


Boolean RockEnter::ignore_walls()
{
   return True;
}


Boolean RockEnter::collidable()
{
  return False;
}


int RockEnter::get_drawing_level()
{
  return 2;
}


void RockEnter::act() 
{
  if (!onTarget) {
     Pos next = get_area_next().get_middle();
//     assert(here.x == target.x);
     if (next.y  >= target.y) {
        onTarget = True;
        frame = 1;
        set_vel_next(0);
        set_middle_next(target);
     }
  } else {
    frame++;
    if (frame == 4) {
       kill_self();
       LocatorP l = get_locator();
       WorldP w = get_world();
       PhysicalP rockMan = new RockMan(w, l, 
		target - 0.5 * RockMan::get_size_max());
       assert(rockMan);
       l->add(rockMan);
       if (get_intel()) {
          rockMan->set_intel(get_intel());
          rockMan->set_dont_collide(get_dont_collide());
          this->set_intel(NULL);
       }
    } else {
      set_frame_next(frame);
    }
  }
  NewAnimated::act();
}





KShield::KShield(WorldP w,LocatorP l,const Pos &pos) 
     : Shield(context,xdata,w,l,pos) 
{}



void KShield::use(PhysicalP p)
{
  stats.add_use();
  Shield::use(p);
}



ProtectionP KShield::create_protection(const Area &area)
{
  ProtectionP pr = new KProtection(get_world(),get_locator(),area);
  assert(pr);
  return pr;
}



Stats KShield::stats;




RShield::RShield(WorldP w,LocatorP l,const Pos &pos) 
     : Shield(context,xdata,w,l,pos) 
{}



void RShield::use(PhysicalP p)
{
  stats.add_use();
  Shield::use(p);
}



ProtectionP RShield::create_protection(const Area &area)
{
  ProtectionP pr = new RProtection(get_world(),get_locator(),area);
  assert(pr);
  return pr;
}



Stats RShield::stats;





KProtection::KProtection(WorldP w,LocatorP l,const Area &area)
  : Protection(context,xdata,w,l,area)
{
  k = K_PROTECTION_K;
}



int KProtection::corporeal_defend(int damage)
{
  assert(k >= 0);

  int protect_max = K_THRESHOLD;
  int returnval;

  if (protect_max > k)
     protect_max = k;

  if (damage < protect_max) {
     k -= damage;
     returnval = 0;
  } else {
     k -= protect_max;
     returnval - protect_max;
  }
  if (k <= 0) {
     kill_self();
  }
  return returnval;
}



int KProtection::heat_defend(int heat, Boolean)
{
  assert(k >= 0);
  return heat;
}



RProtection::RProtection(WorldP w,LocatorP l,const Area &area)
  : Protection(context,xdata,w,l,area)
{
  r = R_PROTECTION_PERCENT;
  r_h = R_PROTECTION_PERCENT;
}


int RProtection::corporeal_defend(int damage)
{
   int returnval;
   returnval = ((damage * 100) - (damage * r)) / 100;
   return returnval;
}



int RProtection::heat_defend(int heat, Boolean)
{
   int returnval;
   returnval = ((heat * 100) - (heat * r)) / 100;
   return returnval;
}



/**********Human Torch********/

JetPack::JetPack(WorldP w, LocatorP l, const Pos &pos)
   : AutoUse(context, xdata, w, l, pos)
{}


void JetPack::use(PhysicalP user) {
   assert(user->is_creature());

   MovingP foo = (MovingP)user;
//   Stance foodir = foo->get_stance_next();

   Vel boost;
   boost.dy = -7;
   boost.dx = 0;

   foo->set_vel_next(foo->get_vel_next() + boost);
}


Stats JetPack::stats;



Kenny::Kenny(WorldP w,LocatorP l,const Pos &rawPos)
     : Creature(creatureContext,creatureXdata,w,l,rawPos),
       Grounded(groundedContext,groundedXdata),
       Suicide(suicideContext,suicideXdata)
{}



Boolean Kenny::is_suicide()
{
  return True;
}



void Kenny::act()
{
  Grounded::_act(); 
  Suicide::_act();
  Creature::act();
}
 
      
 
void Kenny::die()
{
  int cc;
  ostrstream msg;
  LocatorP locator = get_locator();

  if (!get_quiet_death()) {
    stats.add_death(get_birth_time());
    for (cc = 0; cc<10; cc++)
       bleed();
    msg << "OH MY GOD, THEY KILLED KENNY!" << endl << "..YOU BASTARDS!!"
        << ends;
    locator->message_enq(msg.str());
  }
  
  Grounded::die();
}
 
      
Stats Kenny::stats;



SwastikaRang::SwastikaRang(WorldP w, LocatorP l, const Pos &p, 
		const Id &shooter, Dir d)
	: Shot(context, xdata, w, l, p, shooter, d)
{
   stats.add_creation();
   returning = False;
   returningbegun = False;
   returnTimer.set(BOOMERANG_TIME);
}


Boolean SwastikaRang::die_on_impact()
{
   return returningbegun;
}


void SwastikaRang::act()
{
   returnTimer.clock();
   if (returning)
      returningbegun = True;
   if ((returnTimer.ready() || hit_wall()) && !returning) {
      returning = True;
      Vel vel;
      vel = get_vel();
      vel.dx = -vel.dx;
      vel.dy = -vel.dy;
      set_vel_next(vel);
      returnTimer.set(BOOMERANG_TIME);
   } else if (returnTimer.ready()) {
      kill_self();
   }
   Shot::act();
}


void SwastikaRang::collide(PhysicalP other) 
{
   if(!returning && !other->is_shot()) {
      if (other->get_id() != get_shooter()) {
         returning = True;
         Vel vel = get_vel();
         vel.dx = -vel.dx;
         vel.dy = -vel.dy;
         set_vel_next(vel);
        returnTimer.set(BOOMERANG_TIME);
      }
   }
   Shot::collide(other);
}



Stats SwastikaRang::stats;



SwastikaRangs::SwastikaRangs(WorldP w,LocatorP l,const Pos &p) 
: Gun(context,xdata,w,l,p) 
{}



Size SwastikaRangs::get_shot_size(Dir dir)
{
  return SwastikaRang::get_size();
}



PhysicalP SwastikaRangs::create_shot(PhysicalP shooter,WorldP world,
				LocatorP locator,const Pos &pos,Dir dir)
{
  PhysicalP shot = new SwastikaRang(world,locator,pos,shooter->get_id(),dir);
  assert (shot);
  return shot;
}

